module(..., package.seeall)

--====================================================================--
-- LOADER CLASS
--====================================================================--
--[[

 - Version: 1.0
 - Made by Ricardo Rauber Pereira @ 2011
 - Blog: http://rauberlabs.blogspot.com/
 - Mail: ricardorauber@gmail.com

******************
 - LICENSE
******************
 
 - This class is free to use, feel free to change it but please send new versions
   or new features to me and help us to make it better!

******************
 - HISTORY
******************

 - 27-APR-2011 - Ricardo Rauber - Created

******************
 - INFORMATION
******************

 - This class is for expert users, please take a look on the table below.

 |------------------|-------------------------|----------------------------------------------------|
 | FUNCTION         | RETURNS                 | DESCRIPTION                                        |
 |------------------|-------------------------|----------------------------------------------------|
 | showDebug        | void                    | Show debug messages at Terminal when debug = true. |
 | showError        | void                    | Show errors at Terminal and Device Alert.          |
 | callFunction     | function result / false | Execute a function from a module or an object.     |
 | moduleExists     | boolean                 | Verify is the module was loaded.                   |
 | loadModule       | boolean                 | Require a module.                                  |
 | unloadModule     | boolean                 | Remove the module from package.loaded table.       |
 | imageExists      | boolean                 | Verify is the image was loaded.                    |
 | loadImage        | boolean                 | Open an image.                                     |
 | unloadImage      | boolean                 | Remove the image from the tabImages table.         |
 | audioExists      | boolean                 | Verify is the audio was loaded.                    |
 | loadAudio        | boolean                 | Open an audio.                                     |
 | unloadAudio      | boolean                 | Remove the audio from the tabAudio table.          |
 | playAudio        | boolean                 | Play an audio.                                     |
 | timerExists      | boolean                 | Verify is the timer was created.                   |
 | createTimer      | boolean                 | Create a new timer.                                |
 | cancelTimer      | boolean                 | Stop a timer.                                      |
 | newObject        | table / false           | Create an object from an external module.          |
 | newGroup         | display group / false   | Create a display group from an external module.    |
 |------------------|-------------------------|----------------------------------------------------|

--]]

--====================================================================--
-- VARIABLES
--====================================================================--

debug = false
--
local garbageDelay = 200
--
local tabImages = {}
local tabAudio  = {}
local tabTimers = {}

--====================================================================--
-- DEBUG
--====================================================================--

function showDebug ( debugMessage )
	if debug then
		local str = "Loader DEBUG: " .. tostring(debugMessage)
		print ( str )
	end
end

--====================================================================--
-- SHOW ERRORS
--====================================================================--

function showError ( errorMessage )
	local str = "Loader ERROR: " .. tostring(errorMessage)
	local function onComplete ( event )
		print ()
		print ( "-----------------------" )
		print ( str )
		print ( "-----------------------" )
		error ()
	end
	local alert = native.showAlert( "Loader Class - ERROR", str, { "OK" }, onComplete )
end

--====================================================================--
-- GARBAGE COLLECTOR
--====================================================================--

function getGarbageDelay ( )
	return garbageDelay
end
--
function setGarbageDelay ( delay )
	if type( delay ) == "number" then
		garbageDelay = delay
	end
end
--
function garbageCollect ( event )
	collectgarbage( "collect" )
	showDebug ( "garbageCollect - Cleaning memory: " .. collectgarbage("count") )
end

--====================================================================--
-- IS DISPLAY OBJECT ?
--====================================================================--

local coronaMetaTable = getmetatable(display.getCurrentStage())
--
local isDisplayObject = function(aDisplayObject)
	return (type(aDisplayObject) == "table" and getmetatable(aDisplayObject) == coronaMetaTable)
end

--====================================================================--
-- CALL FUNCTION
--====================================================================--

function callFunction ( moduleName, functionName, params )

	-----------------------
	-- Test Parameters
	-----------------------

	if type( functionName ) ~= "string" then
		showError ( "callFunction - The function name must be a string. functionName = " .. tostring(functionName) )
		return false
	end
	
	-----------------------
	-- Verify if it's a module name or an object
	-----------------------
	
	local moduleHandler
	--
	if type( moduleName ) == "string" then
		showDebug ( "callFunction - '" .. tostring(moduleName) .. "' is a module name." )
		moduleHandler = package.loaded[moduleName]
	elseif type( moduleName ) == "table" then
		showDebug ( "callFunction - '" .. tostring(moduleName) .. "' is an object." )
		moduleHandler = moduleName
	elseif type ( moduleName ) == "function" then
		showDebug ( "callFunction - '" .. tostring(functionName) .. "' is a function." )
		moduleHandler = moduleName
	else
		showError ( "callFunction - The module name must be a string, function or an object. moduleName = " .. tostring(moduleName) )
		return false
	end
	--
	if not moduleHandler then
		showError ( "callFunction - Module/Function/Object '" .. tostring(moduleName) .. "' was not found." )
		return false
	end

	-----------------------
	-- Search function
	-----------------------
	
	showDebug ( "callFunction - Searching " .. tostring(moduleName) .. "." .. tostring(functionName) .. "( params )." )
	--
	local fncExec, fncResult
	--
	if type ( moduleHandler ) == "function" then
		fncExec = moduleHandler
	else
		for k,v in pairs( moduleHandler ) do
			if k == tostring(functionName) and type(v) == "function" then
				fncExec = v
			end
		end
	end
	
	-----------------------
	-- Execute
	-----------------------
		
	if fncExec then
		showDebug ( "callFunction - Executing " .. tostring(moduleName) .. "." .. tostring(functionName) .. "( params )." )
		--
		local handler
		handler, fncResult = pcall ( fncExec, params )
		--
		if not handler then
			showError ( "callFunction - Failed to execute " .. tostring(functionName) .. "( params ) function on '" .. tostring(moduleName) .. "'." )
			return false
		end
	else
		showError ( "callFunction - Function " .. tostring(functionName) .. "( params ) was not found on module '" .. tostring(moduleName) .. "'." )
		return false
	end
	
	-----------------------
	-- Return
	-----------------------
	
	showDebug ( "callFunction - Function " .. tostring(functionName) .. "( params ) executed on '" .. tostring(moduleName) .. "'." )
	return fncResult
	
end

--====================================================================--        
-- MODULE EXISTS?
--====================================================================--

function moduleExists ( moduleName )

	-----------------------
	-- Test Parameters
	-----------------------

	if type( moduleName ) ~= "string" then
		showError ( "moduleExists - The image name must be a string. moduleName = " .. tostring( moduleName ) )
		return false
	end
	
	-----------------------
	-- Test if it was loaded
	-----------------------
	
	if type( package.loaded[moduleName] ) ~= "nil" then
		return true
	else
		return false
	end
	
end

--====================================================================--
-- LOAD MODULES
--====================================================================--

function loadModule ( moduleName )

	-----------------------
	-- Test Parameters
	-----------------------

	if type( moduleName ) ~= "string" then
		showError ( "loadModule - The module name must be a string. moduleName = " .. tostring(moduleName) )
		return false
	end
	--
	showDebug ( "loadModule - Loading module '" .. moduleName .. "'." )
	
	-----------------------
	-- Test if it's already loaded
	-----------------------
	
	if type( package.loaded[moduleName] ) == "nil" then
	
		-----------------------
		-- Try to load
		-----------------------
		
		if not pcall ( require, moduleName ) then
			showError ( "loadModule - Failed to load module '" .. moduleName .. "' - Please check if the file exists and it is correct." )
			return false
		end
		
	else
		showDebug ( "loadModule - Module '" .. moduleName .. "' is already loaded." )
		return true
	end
	
	-----------------------
	-- Return
	-----------------------
	
	showDebug ( "loadModule - Module '" .. moduleName .. "' succesfully loaded." )
	return true
	
end

--====================================================================--
-- UNLOAD MODULES
--====================================================================--

function unloadModule ( moduleName )

	-----------------------
	-- Test Parameters
	-----------------------

	if type( moduleName ) ~= "string" then
		showError ( "unloadModule - The module name must be a string. moduleName = " .. tostring(moduleName) )
		return false
	end
	--
	showDebug ( "unloadModule - Unloading module '" .. moduleName .. "'." )
	
	-----------------------
	-- Test if it was loaded
	-----------------------
	
	if type( package.loaded[moduleName] ) == "nil" then
		showDebug ( "unloadModule - Module '" .. moduleName .. "' was not loaded." )
		return true
	end
	
	-----------------------
	-- Remove module
	-----------------------
	
	package.loaded[moduleName] = nil
	
	-----------------------
	-- Clean Memory
	-----------------------
	
	timer.performWithDelay( garbageDelay, garbageCollect )
	
	-----------------------
	-- Return
	-----------------------
	
	showDebug ( "unloadModule - Module '" .. moduleName .. "' succesfully unloaded." )
	return true
	
end

--====================================================================--        
-- IMAGE EXISTS?
--====================================================================--

function imageExists ( imageName )

	-----------------------
	-- Test Parameters
	-----------------------

	if type( imageName ) ~= "string" then
		showError ( "imageExists - The image name must be a string. imageName = " .. tostring( imageName ) )
		return false
	end
	
	-----------------------
	-- Test if it was loaded
	-----------------------
	
	if type( tabImages[imageName] ) ~= "nil" then
		return true
	else
		return false
	end
	
end

--====================================================================--
-- LOAD IMAGES
--====================================================================--

function loadImage ( imageName, imageWidth, imageHeight )

	-----------------------
	-- Test Parameters
	-----------------------

	if type( imageName ) ~= "string" then
		showError ( "loadImage - The image name must be a string. imageName = " .. tostring(imageName) )
		return false
	end
	--
	showDebug ( "loadImage - Loading image '" .. imageName .. "'." )
	
	-----------------------
	-- Test if it's already loaded
	-----------------------
	
	if imageExists ( imageName ) then
		showDebug ( "loadImage - Image '" .. imageName .. "' is already loaded." )
		return true
	end
	
	-----------------------
	-- Try to load it
	-----------------------
		
	local handler, img
	--
	if type(imageWidth) == "number" and type(imageHeight) == "number" then
		handler, img = pcall ( display.newImageRect, imageName, imageWidth, imageHeight )
	else
		handler, img = pcall ( display.newImage, imageName )
	end
	--
	if not handler
	or type(img) == "nil" then
		showError ( "loadImage - Failed to load image '" .. imageName .. "' - Please check if the file exists and the name is correct." )
		return false
	end
		
	-----------------------
	-- Put it on the table
	-----------------------
		
	tabImages[imageName] = img
	tabImages[imageName].isVisible = false
	
	-----------------------
	-- Return
	-----------------------
	
	showDebug ( "loadImage - Image '" .. imageName .. "' succesfully loaded." )
	showDebug ( "loadImage - Texture: " .. system.getInfo( "textureMemoryUsed" ) )
	return true
	
end

--====================================================================--
-- UNLOAD IMAGES
--====================================================================--

function unloadImage ( imageName )

	-----------------------
	-- Test Parameters
	-----------------------

	if type( imageName ) ~= "string" then
		showError ( "unloadImage - The image name must be a string. imageName = " .. tostring(imageName) )
		return false
	end
	--
	showDebug ( "unloadImage - Unloading image '" .. imageName .. "'." )
	
	-----------------------
	-- Test if it was loaded
	-----------------------
	
	if not imageExists ( imageName ) then
		showDebug ( "unloadImage - Image '" .. imageName .. "' was not loaded by loader.loadImage() or not loaded at all." )
		return true
	end
	
	-----------------------
	-- Remove image
	-----------------------
	
	tabImages[imageName]:removeSelf()
	tabImages[imageName] = nil
	
	-----------------------
	-- Clean memory
	-----------------------
	
	timer.performWithDelay( garbageDelay, garbageCollect )
	
	-----------------------
	-- Return
	-----------------------
	
	showDebug ( "unloadImage - Image '" .. imageName .. "' succesfully unloaded." )
	showDebug ( "unloadImage - Texture: " .. system.getInfo( "textureMemoryUsed" ) )
	return true
	
end

--====================================================================--        
-- AUDIO EXISTS?
--====================================================================--

function audioExists ( audioName )

	-----------------------
	-- Test Parameters
	-----------------------

	if type( audioName ) ~= "string" then
		showError ( "audioExists - The audio name must be a string. audioName = " .. tostring( audioName ) )
		return false
	end
	
	-----------------------
	-- Test if it was loaded
	-----------------------
	
	if type( tabAudio[audioName] ) ~= "nil" then
		return true
	else
		return false
	end
	
end

--====================================================================--
-- LOAD AUDIO
--====================================================================--

function loadAudio ( audioName )

	-----------------------
	-- Test Parameters
	-----------------------

	if type( audioName ) ~= "string" then
		showError ( "loadAudio - The audio name must be a string. audioName = " .. tostring(audioName) )
		return false
	end
	--
	showDebug ( "loadAudio - Loading Audio '" .. audioName .. "'." )
	
	-----------------------
	-- Test if it's already loaded
	-----------------------
	
	if audioExists ( audioName ) then
		showDebug ( "loadAudio - Audio '" .. audioName .. "' is already loaded." )
		return true
	end
	
	-----------------------
	-- Try to load it
	-----------------------
		
	local handler, aud = pcall ( audio.loadSound, audioName )
	--
	if not handler
	or type(aud) == "nil" then
		showError ( "loadAudio - Failed to load audio '" .. audioName .. "' - Please check if the file exists and the name is correct." )
		return false
	end
		
	-----------------------
	-- Put it on the table
	-----------------------
		
	tabAudio[audioName] = aud
	
	-----------------------
	-- Return
	-----------------------
	
	showDebug ( "loadAudio - Audio '" .. audioName .. "' succesfully loaded." )
	return true
	
end

--====================================================================--
-- UNLOAD AUDIO
--====================================================================--

function unloadAudio ( audioName )

	-----------------------
	-- Test Parameters
	-----------------------

	if type( audioName ) ~= "string" then
		showError ( "unloadAudio - The audio name must be a string. audioName = " .. tostring(audioName) )
		return false
	end
	--
	showDebug ( "unloadAudio - Unloading audio '" .. audioName .. "'." )
	
	-----------------------
	-- Test if it was loaded
	-----------------------
	
	if not audioExists ( audioName ) then
		showDebug ( "unloadAudio - Audio '" .. audioName .. "' was not loaded by loader.loadAudio() or not loaded at all." )
		return true
	end
	
	-----------------------
	-- Remove audio
	-----------------------
	
	audio.dispose( tabAudio[audioName] )
	tabAudio[audioName] = nil
	
	-----------------------
	-- Clean memory
	-----------------------
	
	timer.performWithDelay( garbageDelay, garbageCollect )
	
	-----------------------
	-- Return
	-----------------------
	
	showDebug ( "unloadAudio - Audio '" .. audioName .. "' succesfully unloaded." )
	return true
	
end

--====================================================================--
-- PLAY AUDIO
--====================================================================--

function playAudio ( audioName, params )

	-----------------------
	-- Test Parameters
	-----------------------

	if type( audioName ) ~= "string" then
		showError ( "playAudio - The audio name must be a string. audioName = " .. tostring(audioName) )
		return false
	end
	--
	showDebug ( "playAudio - Going to play audio '" .. audioName .. "'." )
	
	-----------------------
	-- Test if it was loaded
	-----------------------
	
	if not audioExists ( audioName ) then
		showDebug ( "playAudio - Audio '" .. audioName .. "' was not loaded by loader.loadAudio() or not loaded at all." )
		return true
	end
	
	-----------------------
	-- Play audio
	-----------------------
	
	local handler, soundChannel
	handler, soundChannel = pcall ( audio.play, tabAudio[audioName], params )
	--
	if not handler then
		showError ( "playAudio - The audio '" .. audioName .. "' could not be played." )
		return false
	end
	
	-----------------------
	-- Return
	-----------------------
	
	showDebug ( "playAudio - Audio '" .. audioName .. "' succesfully played." )
	return true
	
end

--====================================================================--        
-- TIMER EXISTS?
--====================================================================--

function timerExists ( timerName )

	-----------------------
	-- Test Parameters
	-----------------------

	if type( timerName ) ~= "string" then
		showError ( "timerExists - The timer name must be a string. timerName = " .. tostring( timerName ) )
		return false
	end
	
	-----------------------
	-- Test if it was loaded
	-----------------------
	
	if type( tabTimers[timerName] ) ~= "nil" then
		return true
	else
		return false
	end
	
end

--====================================================================--        
-- CREATE TIMER
--====================================================================--

function createTimer ( timerName, params )

	-----------------------
	-- Test Parameters
	-----------------------

	if type( timerName ) ~= "string" then
		showError ( "createTimer - The timer name must be a string. timerName = " .. tostring( timerName ) )
		return false
	end
	--
	if type( params.delay ) ~= "number" then
		showError ( "createTimer - You have to inform the delay for timer " .. tostring( timerName ) .. "." )
		return false
	end
	--
	if type( params.func ) ~= "function" then
		showError ( "createTimer - You have to inform the function for timer " .. tostring( timerName ) .. "." )
		return false
	end
	--
	local vRepeat = 1
	if type( params.rep ) == "number" then
		vRepeat = params.rep
	end
	--
	showDebug ( "createTimer - Creating timer'" .. timerName .. "'." )
	
	-----------------------
	-- Test if it was created
	-----------------------
	
	if timerExists ( timerName ) then
		showError ( "createTimer - Timer '" .. timerName .. "' already exists." )
		return false
	end
	
	-----------------------
	-- Create timer
	-----------------------
	
	local handler
	handler, tabTimers[timerName] = pcall( timer.performWithDelay, params.delay, params.func, vRepeat)
	--
	if not handler then
		showError ( "createTimer - Timer '" .. timerName .. "' could not be created." )
		return false
	end
	
	-----------------------
	-- Return
	-----------------------
	
	showDebug ( "createTimer - Timer '" .. timerName .. "' was succesfully created." )
	return true

end

--====================================================================--        
-- CANCEL TIMER
--====================================================================--

function cancelTimer ( timerName )

	-----------------------
	-- Test Parameters
	-----------------------

	if type( timerName ) ~= "string" then
		showError ( "cancelTimer - The timer name must be a string. timerName = " .. tostring( timerName ) )
		return false
	end
	--
	showDebug ( "cancelTimer - Canceling timer'" .. timerName .. "'." )
	
	-----------------------
	-- Test if it was loaded
	-----------------------
	
	if not timerExists ( timerName ) then
		showDebug ( "cancelTimer - Timer '" .. timerName .. "' doesn't exists." )
		return true
	end
	
	-----------------------
	-- Cancel timer
	-----------------------
	
	timer.cancel( tabTimers[timerName] )
	tabTimers[timerName] = nil
	
	-----------------------
	-- Clean memory
	-----------------------
	
	timer.performWithDelay( garbageDelay, garbageCollect )
	
	-----------------------
	-- Return
	-----------------------
	
	showDebug ( "cancelTimer - Timer '" .. timerName .. "' was succesfully canceled." )
	return true

end

--====================================================================--
-- NEW OBJECT
--====================================================================--

function newObject ( moduleName, params )

	-----------------------
	-- Test Parameters
	-----------------------

	if type( moduleName ) ~= "string" then
		showError ( "newObject - The module name must be a string. moduleName = " .. tostring(moduleName) )
		return false
	end
	--
	if string.lower( moduleName ) == "main"
	or string.lower( moduleName ) == "config"
	then
		showError ( "newObject - You cannot load " .. moduleName .. ".lua as a module." )
		return false
	end
	--
	showDebug ( "newObject - Creating object '" .. moduleName .. "'." )
	
	-----------------------
	-- Load choosed module
	-----------------------
	
	if not loadModule ( moduleName ) then
		return false
	end
	
	-----------------------
	-- Search new() function
	-----------------------
	
	local newObject = callFunction ( moduleName, "new", params )
	
	-----------------------
	-- Verify the object type
	-----------------------
	
	if type ( newObject ) ~= "table" then
		showError ( "newObject - Function new() returned '" .. type ( newObject ) .."' instead of 'table' on module '" .. moduleName .. "'" )
		return false
	end
	
	-----------------------
	-- Return
	-----------------------
	
	showDebug ( "newObject - Object '" .. moduleName .. "' was succesfully created." )
	return newObject

end
	
--====================================================================--
-- NEW GROUP
--====================================================================--

function newGroup ( moduleName, params )

	-----------------------
	-- Load object
	-----------------------
	
	local screenGroup = newObject ( moduleName, params )
	if not screenGroup then
		return false
	end
	
	-----------------------
	-- Test if it is a group
	-----------------------
	
	showDebug ( "newGroup - Testing if object '" .. moduleName .. "' is a display group." )
	if not isDisplayObject( screenGroup ) then 
		showError ( "newGroup - The new() function must return a display group on module '" .. moduleName .. "'" )
		return false
	end
	
	-----------------------
	-- Return
	-----------------------
	
	showDebug ( "newGroup - Object '" .. moduleName .. "' is OK." )
	return screenGroup

end